/*****************************************************************************
 * F1 Racing diteur/simulateur
 * Fichier: widgettrack.cpp
 *
 * (c) Leroux philippe, 2003
 *
 *
 * This program is distributed under the terms of the GNU General Public
 *
 * License:
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****************************************************************************/
#include "widgettrack.h"

#include <qpainter.h>
#include <qfile.h>
#include <qtextstream.h>

#include "divers/ioerrors.h"

#include "pixel/listshapepixel.h"
#include "pixel/outputpixel.h"
#include "trackpixmap.h"

#include <algorithm>	// pour swap

namespace 
{
	const int he = (Dim::size * Dim::maxY) + (Dim::maxY * Dim::gap)
																+ Dim::gap*2;
	const int wi = (Dim::size * Dim::maxX) + (Dim::maxX * Dim::gap);
	
	const char	asciiPiste	= ' ';

	inline const ShapePixel& shapeOf( const QColor& color )
	{
		return ListShapePixel::instance()[ListShapePixel::id(color)];
	}
};

WidgetTrack::WidgetTrack( int widthPen, QWidget* parent, const char* name )
:	 QWidget( parent, name )
	,state_(Empty)
	,widthPen_(widthPen)
	,minimumSize_(wi,he)
	,xo_(-1)
	,defaultShape_(ListShapePixel::instance()[Vert])
{
	const ShapePixel& shapeTrack = ListShapePixel::instance()[Gris];
	//	----
	//	Initialisation du vecteur de pixels:
	//		- attribution de son n d'indice (pratique)
	//		- des coordonnes sur la piste.
	//		- des coordonnes sur l'cran.
	//		- d'une couleur par dfaut.
	//	----
	for (	int iPixel = 0, y = 0, coordY = Dim::gap ;
			y < Dim::maxY ;
			++y, coordY += Dim::size + Dim::gap )
	{
		for (	int x = 0, coordX = Dim::gap ;
				x < Dim::maxX ;
				++x, coordX += Dim::size + Dim::gap, ++iPixel )
		{
			vpixels_[iPixel].setNum( iPixel );
			vpixels_[iPixel].setCoords( coordX, coordY );
			vpixels_[iPixel].set( x, y );
			vpixels_[iPixel].setShape(
						Dim::inStarterArea(x,y) ? shapeTrack : defaultShape_ );
		}
	}
}

bool WidgetTrack::load( const QString& name, bool clean )
{
	state_ = Empty;
	QFile file( name );
	if ( ! file.open( IO_ReadOnly ) )
	{
		emit warning( true, "Impossible d'ouvrir: " + name 
							+ '\n' + getIoError(file.status()) );
		return false;
	}

	ListShapePixel& listShapePixel = ListShapePixel::instance();

	int		countCorrections = 0;
	QString	listMsg;

	QTextStream	stream( &file );
	QString		line;
	int			iPixel = 0;
	int			iline = 0;

	while ( ! stream.atEnd() )
	{
		line = stream.readLine();
		const int lenLine = static_cast<int>(line.length());
		if ( lenLine != Dim::maxX )
		{
			QString msg;
			QTextOStream( &msg )
				<< name << ":\nLigne " << iline+1 << " trop courte "
				<< line.length() << " symboles au lieu de "
				<<  Dim::maxX << ".\n";
			listMsg += msg;
		}

		for ( int x = 0; x < Dim::maxX; ++x, ++iPixel )
		{
			if ( x >= lenLine || ! listShapePixel.find( line[x].latin1() ) )
			{
				++countCorrections;
				vpixels_[iPixel].setShape( defaultShape_ );
			}
			else
				vpixels_[iPixel].setShape( listShapePixel.shape() );
		}

		++iline;
	}

	if ( iline != Dim::maxY )
	{
		// On attribue la forme par dfaut aux lignes absentes.
		do 
		{
			++countCorrections;
			vpixels_[iPixel].setShape( defaultShape_ );
		} while ( ++iPixel < Dim::maxPixels );

		QString msg;
		QTextOStream( &msg ) 
						<< name << ":\nIl n'y a que " << iline << " lignes: "
						<< Dim::maxY << " attendues.\n";
		listMsg += msg;
		state_ = Changed;
	}
	
	if ( countCorrections != 0 )
	{
		QString msg;
		QTextOStream( &msg )
			<< "Il y a eu " << countCorrections
			<< " corrections de pixels transforms en pixels vert.";
		listMsg += msg;
		state_ = Changed;
	}
	
	if ( ! listMsg.isEmpty() )
		emit warning( false, listMsg );

	if ( state_ == Empty )
	{
		state_ = Saved;
		emit message( "Piste charge." );
	}

	repaint( clean );
	return true;
}

bool WidgetTrack::save( const QString& name )
{
	QFile file( name );
	if ( file.open( IO_WriteOnly ) )
	{
		ListShapePixel& listShapePixel = ListShapePixel::instance();

		QTextStream stream( &file );
		for ( int iPixel = 0; iPixel < Dim::maxPixels; ++iPixel )
		{
			listShapePixel.find( vpixels_[iPixel].color() );
			stream << listShapePixel.ascii();
			if ( (iPixel+1) % Dim::maxX == 0 )
				stream << '\n';
		}
		file.close();
	}

	if ( file.status() == IO_Ok )
	{
		emit message( "Piste sauve dans " + name );
		state_ = Saved;
		return true;
	}

	emit warning( true,
			"Erreur enregistrement de la piste " + name + " :\n\n"
			+ getIoError( file.status() ) );
	return false;
}

void WidgetTrack::select( const QColor& c, int x0, int y0, int x1, int y1 )
{
	if ( xo_ != -1 )
		cleanSelect();

	QPainter painter( this );
	painter.setBrush( c );
	QPen p = painter.pen();
	p.setColor( c );
	p.setWidth( Dim::gap*2 );
	painter.setPen( p );

	xo_ = x0 + Dim::size / 2;
	yo_ = y0 + Dim::size / 2;
	xd_ = x1 + Dim::size / 2;
	yd_ = y1 + Dim::size / 2;
	if ( xo_ != xd_ || yo_ != yd_ )
	{
		QRect rect;
		rect.setCoords( xo_, yo_, xd_, yd_ ); 
		painter.drawRect( rect );
	}
	else
		painter.drawLine( xo_, yo_, xd_, yd_ );
}

void WidgetTrack::cleanSelect()
{
	if ( xo_ > xd_ ) std::swap( xo_, xd_ );
	if ( yo_ > yd_ ) std::swap( yo_, yd_ );
	QRect rect2repaint;
	rect2repaint.setCoords( xo_-Dim::size / 2, yo_-Dim::size / 2 , xd_, yd_ );
	repaint( rect2repaint, true );
	xo_ = -1;
}

TrackPixmap* WidgetTrack::fill( const QColor& c, Pixel* pixelO, Pixel* pixelF )
{
	int coordXDebut	= pixelO->coordX();
	int coordXFin	= pixelF->coordX();
	int coordYDebut	= pixelO->coordY();
	int coordYFin	= pixelF->coordY();

	if ( coordXDebut > coordXFin )	std::swap( coordXDebut, coordXFin );
	if ( coordYDebut > coordYFin )	std::swap( coordYDebut, coordYFin );

	TrackPixmap* imageBefore =
			repaintArea( ::shapeOf(c),
							coordXDebut, coordYDebut, coordXFin, coordYFin );

	setChanged();
	return imageBefore;
}

#define valide( x, y )											\
	(	 vpixels_[Dim::numPixel(x,y)].ascii() == asciiPiste		\
	  || vpixels_[Dim::numPixel(x,y)].color() == driverColor	\
	  || vpixels_[Dim::numPixel(x,y)].color() == traceColor		\
	)

int WidgetTrack::onTrack( const QColor& driverColor, const QColor& traceColor,
												int x, int y, int xf, int yf )
{
	//	---------------------------------------------------
	//	Adaptation de l'algo de L. Rivire, adapter par CLX
	//	---------------------------------------------------
	if ( ! valide( x, y ) )
		return Dim::numPixel(x,y);

	if ( x == xf && y == yf )
		return -1;

	int dx;
	int xi;
	if ( x < xf )
	{
		xi = 1;
		dx = xf - x;
	}
	else
	{
		xi = -1;
		dx = x - xf;
	}

	int dy;
	int yi;
	if ( y < yf )
	{
		yi = 1;
		dy = yf - y;
	}
	else
	{
		yi = -1;
		dy = y - yf;
	}

	bool ok = true;

	if ( dx > dy )
	{
		int total = 0;
		do
		{
			x		+= xi;
			total	+= 2*dy;
			ok		= (total == dx) ? valide( x, y ) : false;

			if ( total >= dx )
			{
				y		+= yi;
				total	-= 2*dx;
			}
			
			ok |= valide( x, y );
		} while ( x != xf && ok );
	}
	else
	{
		int total = 0;
		do
		{
			y		+= yi;
			total	+= 2*dx;
			ok		= (total == dy) ? valide( x, y ) : false;

			if ( total >= dy )
			{
				x		+= xi;
				total	-= 2*dy;
			}

			ok |= valide( x, y );
		} while ( y != yf && ok );
	}

	return ok ? -1 : Dim::numPixel(x,y);
}

void WidgetTrack::paintPixel( const QColor& c, int num )
{
	Pixel& pixel = vpixels_[num];
	pixel.setShape( ::shapeOf(c) );

	const QRect rect( pixel.coordX(), pixel.coordY(), Dim::size, Dim::size );
	QPainter painter( this );
	painter.setPen( c );
	painter.setBrush( c );
	painter.drawRect( rect );
}

void WidgetTrack::play( const TrackPixmap& imageBefore, TrackPixmap* redo )
{
	TrackPixmap::Iterator it = imageBefore.iterator();

	assert( *it != NULL ); // il doit y avoir au moins une image avant !

	do
	{
		const TrackPixmap::PixelInfo& pi = **it;
		if ( redo )
			redo->addPixel( pi.num(), vpixels_[pi.num()].shape() );
		vpixels_[pi.num()].setShape( pi.shape() );
		++it;
	} while ( *it != NULL );

	repaint( imageBefore.rect(), true );

	setChanged();
}

bool WidgetTrack::normalizeTrackBorder()
{
	int	rightCol	= 0;
	int	leftCol		= 0;
	int	topLine		= 0;
	int	bottomLine	= 0;

	//	Lignes suprieur et infrieur.
	const int offsetInf = Dim::maxPixels-Dim::maxX;
	for ( int iPixel = 0; iPixel < Dim::maxX; ++iPixel )
	{
		if ( vpixels_[iPixel].ascii() == asciiPiste )
		{
			++topLine;
			vpixels_[iPixel].setShape( defaultShape_ );
		}

		if ( vpixels_[iPixel+offsetInf].ascii() == asciiPiste )
		{
			++bottomLine;
			vpixels_[iPixel+offsetInf].setShape( defaultShape_ );
		}
	}

	//	Lignes droite et gauche.
	for ( int iPixel = Dim::maxX; iPixel <= Dim::maxPixels-2*Dim::maxX;
														iPixel += Dim::maxX )
	{
		if ( vpixels_[iPixel].ascii() == asciiPiste )
		{
			++rightCol;
			vpixels_[iPixel].setShape( defaultShape_ );
		}

		if ( vpixels_[iPixel+Dim::maxX-1].ascii() == asciiPiste )
		{
			++leftCol;
			vpixels_[iPixel+Dim::maxX-1].setShape( defaultShape_ );
		}
	}

	if ( rightCol != 0 || leftCol != 0 || topLine != 0 || bottomLine != 0 )
	{
		repaint( true );
		QString listAno( "Liste des corrections:\n" );
		if ( rightCol != 0 )
			listAno += "\nColonne droite : " + QString::number(rightCol) + " pixels.";
		if ( leftCol != 0 )
			listAno += "\nColonne gauche : " + QString::number(leftCol) + " pixels.";
		if ( topLine != 0 )
			listAno += "\nLigne suprieur: " + QString::number(topLine) + " pixels.";
		if ( bottomLine != 0 )
			listAno += "\nLigne infrieur: " + QString::number(bottomLine) + " pixels.";
	
		listAno += "\n\nRemplacer par des pixels vert.";
		emit warning( false, listAno );

		setChanged();
		return true;
	}

	return false;
}

bool WidgetTrack::checkStarterArea()
{
	//Rectangle dlimit par les points x=20,y=2 et x=59,y=6.
	//C'est au milieu de cette zone que se situe le point de dpart (x=40,y=4).

	const ShapePixel&	shapePiste	= ListShapePixel::instance()[Gris];
	const int			hPixel		= 59 - 20 + 1;
	int					corrections	= 0;
	int					iPixel		= Dim::numPixel(20,2);
	for ( int y = 2; y <= 6; ++y, iPixel += Dim::maxX - hPixel )
	{
		for ( int x = 20; x <= 59; ++x, ++iPixel )
		{
			if ( vpixels_[iPixel].ascii() != asciiPiste )
			{
				++corrections;
				vpixels_[iPixel].setShape( shapePiste );
			}
		}
	}
	
	if ( corrections != 0 )
	{
		const Pixel&	p1 = vpixels_[Dim::numPixel(20,2)];
		const Pixel&	p2 = vpixels_[Dim::numPixel(59,6)];
		const QRect		rect( p1.coordX(), p1.coordY(),
						p2.coordX() + (Dim::gap+Dim::size) - p1.coordX(),
						p2.coordY() + (Dim::gap+Dim::size) - p1.coordY() );

		repaint( rect, true );
		emit warning( false,
						"Il y a eu "+QString::number(corrections)
					+	" corrections dans la zone de dpart [20,2] [59,6]." );

		setChanged();
		return true;
	}

	return false;
}

int WidgetTrack::countColor( const QColor& c )
{
	int count = 0;

	for ( int i = 0; i < Dim::maxPixels; ++i )
		if ( vpixels_[i].color() == c )
			++count;

	return count;
}

void WidgetTrack::paintEvent( QPaintEvent* pe )
{
	const QRect& rpe = pe->rect();
	int leftTopX, leftTopY, rightBottomX, rigthBottomY;
	rpe.coords( &leftTopX, &leftTopY, &rightBottomX, &rigthBottomY );

	QRect rect;
	rect.setCoords(	 leftTopX - (Dim::size+Dim::gap)
					,leftTopY - (Dim::size+Dim::gap)
					,rightBottomX + (Dim::size+Dim::gap)
					,rigthBottomY + (Dim::size+Dim::gap)
					);

	QPainter	painter( this );
	QPen		pen = painter.pen();
	pen.setWidth( widthPen_ );
	painter.setPen( pen );

	char prevAscii = '@';

	for ( int i = 0; i < Dim::maxPixels; ++i )
	{
		const Pixel& pixel = vpixels_[i];
		if ( rect.contains( pixel.coordX(), pixel.coordY() ) )
		{
			if ( prevAscii != pixel.ascii() )
			{
				prevAscii = pixel.ascii();
				painter.setBrush( pixel.color() );
				pen.setColor( pixel.color() );
				painter.setPen( pen );
			}
			painter.drawRect( pixel.coordX(), pixel.coordY(),
														Dim::size, Dim::size );
		}
	}
}

void WidgetTrack::mousePressEvent( QMouseEvent* me )
{
	const int numPixel = Dim::numPixelCoords( me->x(), me->y() );
	if ( Dim::numPixelValid( numPixel ) )
		emit mouseLeftPressPixel( me, vpixels_[numPixel] );
}

void WidgetTrack::mouseMoveEvent( QMouseEvent* me )
{
	const int numPixel = Dim::numPixelCoords( me->x(), me->y() );
	if ( Dim::numPixelValid( numPixel ) )
		emit mouseMouveOnPixel( vpixels_[numPixel] );
}

void WidgetTrack::mouseReleaseEvent( QMouseEvent* me )
{
	const int numPixel = Dim::numPixelCoords( me->x(), me->y() );
	if ( Dim::numPixelValid( numPixel ) )
		emit mouseLeftReleasePixel( me, vpixels_[numPixel] );
}

TrackPixmap* WidgetTrack::repaintArea( const ShapePixel& shapePixel,
											int x0, int y0, int x1, int y1 )
{
	const Pixel&	po		= vpixels_[Dim::numPixelCoords(x0,y0)];
	const Pixel&	pd		= vpixels_[Dim::numPixelCoords(x1,y1)];
	const int		nbCol	= pd.x() - po.x();
	const int		nbLine	= pd.y() - po.y();
	const QRect		rect(	po.coordX(), po.coordY(),
							pd.coordX() + (Dim::gap+Dim::size) - po.coordX(),
							pd.coordY() + (Dim::gap+Dim::size) - po.coordY() );

	TrackPixmap* before = new TrackPixmap( rect );

	int ipixel	= po.num();
	int col		= 0;
	int line	= 0;
	while ( line <= nbLine )
	{
		before->addPixel( ipixel, vpixels_[ipixel].shape() );
		vpixels_[ipixel].setShape( shapePixel );
		if ( ++col > nbCol )
		{
			col = 0;
			++line;
			ipixel += Dim::maxX - nbCol;
		}
		else
			++ipixel;
	}

	repaint( rect, true );
	return before;
}

