/*****************************************************************************
 * F1 Racing diteur/simulateur
 * Fichier: widgettrack.h
 *
 * (c) Leroux philippe, 2003
 *
 *
 * This program is distributed under the terms of the GNU General Public
 *
 * License:
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****************************************************************************/
#ifndef WIDGETTRACK_H
#define WIDGETTRACK_H

//============================================================================
//	WidgetTrack:
//		Gestion de l'affichage d'une piste.
//			- chargement d'une piste.
//			- modification des pixels d'une piste.
//			- enregistrement d'une piste.
//			- test la validiter des bordures de la piste.
//
//============================================================================

#include <qwidget.h>
#include <qvector.h>

#include "pixel/pixel.h"
#include "dimensions.h"

#include <cassert>

class TrackPixmap;
class WidgetTrack : public QWidget
{
Q_OBJECT
public:
	WidgetTrack( int widthPen, QWidget* parent, const char* name );

	enum State { Empty, Changed, Saved };
	State			state() const { return state_; }

	// Charge la piste du fichier 'name'
	bool			load( const QString& name, bool clean=true );
	// Enregiste la piste courante dans le fichier 'name'.
	bool			save( const QString& name );


	//	select et cleanSelect ne modifient pas la piste.
	//	select	:	trace une ligne.
	//	cleanSelect	:	efface la ligne trace par select.
	void			select( const QColor&, int, int, int, int );
	void			select( const QColor&, int numPixel0, int numPixel1 );
	void			cleanSelect();
	void			cancelCleanSelect() { xo_ = -1; }

	//	Colorise avec la couleur 'indexColor' les pixels situs dans la zone
	//	dlimite pour les deux pixels passs en paramtre.
	TrackPixmap*	fill( const QColor&, Pixel*, Pixel* );
	TrackPixmap*	fill( const QColor&, int numPixel0, int numPixel1 );

	//	Dessine l'image 'imageBefore', si 'redo' est non null renvoie
	//	l'ancienne image de la zone de imageBefore.
	void 			play( const TrackPixmap& imageBefore,
													TrackPixmap* redo=NULL );

	//	Les 3 fonctions ne modifie pas la piste : ie state_ n'est pas affect.
	int				onTrack( const QColor& driverColor,
								const QColor& traceColor, int, int, int, int );
	void			paintPixel( const QColor& c, int num );
	int				moveCar( const QColor& c, const QColor& t,
															int of, int to );

	// Vrifie et corrige les bordures de la piste.
	bool			normalizeTrackBorder();
	bool			checkStarterArea();
	//	Compte le nombre de fois que 'c' est utilis par la piste.
	int				countColor( const QColor& c );

	void			setWidthPen( int w )	{ widthPen_ = w; }

	override QSize	minimumSizeHint() const { return minimumSize_; }

signals:
	void warning( bool error, const QString& );
	void message( const QString& );

	//	Sur la slection du boutton de la sourie, envoie le pixel sous
	//	le curseur.
	void mouseLeftPressPixel( QMouseEvent*, Pixel& );
	// Les mouvements de la sourie sont transmis entre un 'press' et un
	// 'release'.
	void mouseMouveOnPixel( Pixel& );
	//	Sur le relachement du boutton de la sourie, envoie le pixel sous
	//	le curseur.
	void mouseLeftReleasePixel( QMouseEvent*, Pixel& );
	
	//	Une modification sur la piste a t faite.
	void signalChanged();

private:
	void setChanged();

	override void paintEvent( QPaintEvent* );
	override void mousePressEvent( QMouseEvent* );
	override void mouseMoveEvent( QMouseEvent* );
    override void mouseReleaseEvent( QMouseEvent* );

	//	Colorise les pixels situs dans la ligne dlimite par les deux
	//	pixels passs en paramtre, renvoie une image de la zone avant
	//	modification.
	TrackPixmap* repaintArea( const ShapePixel&,
											int x0, int y0, int x1, int y1 );

private:
	State		state_;

	int			widthPen_;

	const QSize	minimumSize_;
	Pixel		vpixels_[Dim::maxPixels];	//	Vecteur de tous les pixels
											//	constituant la piste.

	//	Mmorise les coordonnes de la ligne trac, la ligne n'altre pas
	//	la piste.
	int			xo_, yo_;
	int			xd_, yd_;

	const ShapePixel&	defaultShape_;
};



inline TrackPixmap*
WidgetTrack::fill( const QColor& c, int numPixel0, int numPixel1 )
{
	assert( Dim::numPixelValid( numPixel0 ) && Dim::numPixelValid( numPixel1 ) );

	return fill( c, &vpixels_[numPixel0], &vpixels_[numPixel1] );
}

inline void
WidgetTrack::select( const QColor& c, int numPixel0, int numPixel1 )
{
	assert( Dim::numPixelValid( numPixel0 ) && Dim::numPixelValid( numPixel1 ) );

	const Pixel& p1 = vpixels_[numPixel0];
	const Pixel& p2 = vpixels_[numPixel1];
	select( c, p1.coordX(), p1.coordY(), p2.coordX(), p2.coordY() );
}

inline void WidgetTrack::setChanged()
{
	if ( state_ != Changed )
	{
		state_ = Changed;
		emit signalChanged();
	}
}

inline int	WidgetTrack::moveCar( const QColor& c, const QColor& t, 
															int of, int to )
{
	assert( Dim::numPixelValid( of ) && Dim::numPixelValid( to ) );

	const Pixel&	pof	= vpixels_[of];
	const Pixel&	pto	= vpixels_[to];
	const int		ot	= onTrack( c, t, pof.x(), pof.y(), pto.x(), pto.y() );
	paintPixel( c, to );
	return ot;
}

#endif	//WIDGETTRACK_H


